﻿using System;
using System.IO;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using DocumentFormat.OpenXml;
using DocumentFormat.OpenXml.Packaging;
using DocumentFormat.OpenXml.Wordprocessing;

namespace converter.test
{
    [TestClass]
    public sealed class TestFiles
    {
        private TestContext testContextInstance;

        public TestContext TestContext
        {
            get { return testContextInstance; }
            set { testContextInstance = value; }
        }

        private string GetPrefixedName(OpenXmlAttribute attribute)
        {
            return string.Format("{0}:{1}", attribute.Prefix, attribute.LocalName);
        }

        private string GetPrefixedName(OpenXmlElement element)
        {
            return string.Format("{0}:{1}", element.Prefix, element.LocalName);
        }

        // Compare two elements recursively.
        private void CompareElements(string xpath, OpenXmlElement target, OpenXmlElement correct)
        {
            xpath += string.Format("{0}/", GetPrefixedName(correct));

            // Check node names.
            Assert.AreEqual(GetPrefixedName(correct), GetPrefixedName(target),
                "Node: {0}: expected {1}, actual {2}", GetPrefixedName(correct), GetPrefixedName(target));

            // Check content.
            Assert.AreEqual(correct.InnerText, target.InnerText,
                "Content: {0}: expected {1}, actual {2}", xpath, correct.InnerText, target.InnerText);

            // Check attributes.
            Assert.AreEqual(correct.GetAttributes().Count, target.GetAttributes().Count,
                "Number attributes: {0}: expected {1}, actual {2}", xpath, correct.GetAttributes().Count, target.GetAttributes().Count);
            for (int i = 0; i < target.GetAttributes().Count; i++)
            {
                // Attribute name.
                Assert.AreEqual(GetPrefixedName(correct.GetAttributes()[i]), GetPrefixedName(target.GetAttributes()[i]),
                    "Attribute name: {0}: expected {1}, actual {2}", xpath, GetPrefixedName(correct.GetAttributes()[i]), GetPrefixedName(target.GetAttributes()[i]));

                // Content.
                Assert.AreEqual(correct.GetAttributes()[i].Value, target.GetAttributes()[i].Value,
                    "Attribute value: {0}: expected {1}, actual {2}", xpath, correct.GetAttributes()[i].Value, target.GetAttributes()[i].Value);
            }

            // Check children nodes count.
            Assert.AreEqual(correct.ChildElements.Count, target.ChildElements.Count,
                "Number children nodes: {2}: expected {0}, actual {1}", correct.ChildElements.Count, target.ChildElements.Count, xpath);

            // Going deeper.
            for (int i = 0; i < target.ChildElements.Count; i++)
                CompareElements(xpath, target.ChildElements[i], correct.ChildElements[i]);
        }

        // Compare parts of two documents.
        private void CompareParts(OpenXmlPart target, OpenXmlPart correct)
        {
            OpenXmlElement targetElement = target.RootElement;
            OpenXmlElement correctElement = correct.RootElement;
            CompareElements("/", targetElement, correctElement);
        }

        // Compare a target .docx document with its correct version.
        private void CompareDocuments(WordprocessingDocument target, WordprocessingDocument correct)
        {
            // Go through each part of the target document.
            foreach (IdPartPair targetPair in target.Parts)
            {
                OpenXmlPart targetPart = targetPair.OpenXmlPart;
                // Find corresponding part in the correct document.
                IEnumerable<OpenXmlPart> correctParts =
                    from pair in correct.Parts
                    where pair.OpenXmlPart.GetType() == targetPart.GetType()
                    select pair.OpenXmlPart;
                // If the corresponding part is found, compare the parts.
                if (correctParts.Count() > 0)
                    CompareParts(targetPart, correctParts.First());
                else
                    Assert.Fail("The part of type {0} does not exist in the correct document.", targetPart.GetType());
            }
        }

        // Compare a generated report with its correct version.
        private void TestFile(TestReport report)
        {
            TestContext.WriteLine("Comparing: correct {0}, target {1}", report.Correct, report.Target);

            WordprocessingDocument target = WordprocessingDocument.Open(report.Target, false); // open read-only
            WordprocessingDocument correct = WordprocessingDocument.Open(report.Correct, false); // open read-only
            CompareDocuments(target, correct);
        }

        private bool hasErrors = false;

        // Compare all previously generated test reports with their correct versions.
        [TestMethod]
        public void TestAllFiles()
        {
            foreach (TestReport report in Program.Reports)
                try
                {
                    TestFile(report);
                }
                catch (AssertFailedException err)
                {
                    TestContext.WriteLine(err.Message);
                    hasErrors = true;
                }
            if (hasErrors)
                Assert.Fail("Test failed.");
        }
    }
}
